import sys, traceback, os, time

def gettimezone():
    """Return the number of seconds to add to GMT to get local time."""
    t = int(time.time())
    local = time.localtime(t)[:6] + (0, 0, 0)
    universal = time.gmtime(t)[:6] + (0, 0, 0)
    return int(time.mktime(local) - time.mktime(universal))

def mkgmtime(datetime):
    """Given a (y, l, d, h, m, s) tuple in GMT, return the Unix time."""
    daylight = time.localtime()[8:]
    datetime = (datetime + (0, 0, 0, 0, 0))[:8] + daylight
    return time.mktime(datetime) + gettimezone()

def getdate(unixtime, dayoffset=0):
    """Get the GMT date for a given Unix time, offset by a number of days."""
    y, l, d, h, m, s = time.gmtime(unixtime + dayoffset*24*3600)[:6]
    return y, l, d

class Log:
    def __init__(self, path, keepdays=5, listener=None):
        self.path = path
        self.keepdays = keepdays
        self.listener = listener
        self.saved = 0
        self.load()

    def writefile(self, date, entries):
        if not os.path.isdir(self.path):
            os.makedirs(self.path)
        filename = '%04d-%02d-%02d.txt' % date
        filepath = os.path.join(self.path, filename)
        if not entries and not os.path.isfile(filepath):
            return
        file = open(filepath + '.tmp', 'w')
        for start, action in entries:
            y, l, d, h, m, s = time.gmtime(start)[:6]
            action = unicode(action).encode('utf-8')
            file.write('%02d:%02d:%02d %s\n' % (h, m, s, action))
        file.close()
        os.rename(filepath + '.tmp', filepath)

    def readfile(self, date):
        filename = '%04d-%02d-%02d.txt' % date
        filepath = os.path.join(self.path, filename)
        entries = []
        if os.path.isfile(filepath):
            for line in open(filepath):
                time, action = (line + ' ').split(' ', 1)
                action = action.strip().decode('utf-8')
                h, m, s = map(int, time.split(':'))
                start = mkgmtime(date + (h, m, s))
                entries.append((start, action))
        return entries

    def load(self, days=None, endtime=None):
        days = days or self.keepdays
        endtime = endtime or time.time()
        entries = [self.readfile(getdate(endtime, day))
                   for day in range(1 - days, 1)]
        self.entries = sum(entries, [])
        self.saved = 1

    def trim(self, days=None, endtime=None):
        days = days or self.keepdays
        endtime = endtime or time.time()
        dayentries = {}
        for day in range(1 - days, 1):
            dayentries[getdate(endtime, day)] = []
        for start, action in self.entries:
            date = time.gmtime(start)[:3]
            if date in dayentries:
                dayentries[date].append((start, action))
        entries = [dayentries[getdate(endtime, day)]
                   for day in range(1 - days, 1)]
        self.keepdays, self.entries = days, sum(entries, [])
        return dayentries

    def save(self, days=None):
        if not self.saved:
            try:
                dayentries = self.trim(days)
                for date in dayentries:
                    self.writefile(date, dayentries[date])
            except:
                print >>sys.stderr, '\n' + time.asctime() + '\n'
                e, v, t = sys.exc_info()
                traceback.print_exception(e, v, t, file=sys.stderr)
                sys.exit(1)
        self.saved = 1

    def append(self, (start, action)):
        start = start or time.time()
        action = action or ''
        self.entries.append((start, action))
        self.saved = 0
        if self.listener:
            self.listener()

    def pop(self):
        entry = self.entries.pop(-1)
        self.saved = 0
        if self.listener:
            self.listener()
        return entry

    def __getitem__(self, index):
        return self.entries[index]

    def __setitem__(self, index, entry):
        self.entries[index] = entry
        self.saved = 0
        if self.listener:
            self.listener()

    def __getslice__(self, start, end):
        if start < 0:
            start = 0
        if end > len(self.entries):
            end = len(self.entries)
        return self.entries[start:end]

    def __setslice__(self, start, end, entries):
        if start < 0:
            start = 0
        if end > len(self.entries):
            end = len(self.entries)
        self.entries[start:end] = entries
        self.saved = 0
        if self.listener:
            self.listener()

    def __len__(self):
        return len(self.entries)
