from Tkinter import *

def curry(function, *curryargs, **currykw):
    def curried(*args, **kw):
        args = curryargs + args
        kw = dict(currykw.items() + kw.items())
        return function(*args, **kw)
    return curried

class AquaLabel(Label):
    defaults = dict(
        font='{Lucida Grande} 11', borderwidth=1, background='#ffffff',
        foreground='#000000', highlightthickness=0)

    def __init__(self, parent=None, **options):
        for name in self.defaults:
            if name not in options:
                options[name] = self.defaults[name]
        Label.__init__(self, parent, **options)

class AquaEntry(Entry):
    defaults = dict(
        font='{Lucida Grande} 11', borderwidth=0, background='#ffffff',
        editingforeground='#ff4000', foreground='#000000',
        highlightthickness=1, insertontime=333, insertofftime=333,
        insertwidth=2, insertbackground='#808080',
        selectborderwidth=0, selectbackground='#b5d5ff')

    def __init__(self, parent=None, **options):
        for name in self.defaults:
            if name not in options:
                options[name] = self.defaults[name]
        self.command = None
        if 'command' in options:
            self.command = options['command']
            del options['command']
        if not 'highlightbackground' in options:
            options['highlightbackground'] = options['background']
        self.foreground = options['foreground']
        self.editingforeground = options['editingforeground']
        del options['editingforeground']
        Entry.__init__(self, parent, **options)

        self.variable = StringVar()
        self['textvariable'] = self.variable
        self.variable.trace('w', self.change)
        self.bind('<Key-Return>', self.confirm)
        self.bind('<Key-Escape>', self.cancel)
        self.bind('<Key-Up>', self.home)
        self.bind('<Key-Down>', self.end)
        self.bind('<Command-Key-Left>', self.home)
        self.bind('<Command-Key-Right>', self.end)
        self.bind('<KeyRelease>', self.monitor)
        self.bind('<ButtonRelease>', self.monitor)
        self.bind('<FocusIn>', self.selectall)
        self.bind('<Command-Key>a', self.selectall)
        self.bind('<Command-Key>z', self.cancel)
        self.value = ''
        self.changed = 0
        self.state = 0, None

    def getstate(self):
        if self.selection_present():
            selection = self.index('sel.first'), self.index('sel.last')
        else:
            selection = None
        return self.index('insert'), selection

    def setstate(self, (cursor, selection)):
        if selection:
            self.selection_range(*selection)
        else:
            self.selection_clear()
        self.icursor(cursor)

    def focus(self, *args):
        Entry.focus(self)

    def selectall(self, *args):
        self.selection_range(0, 'end')

    def home(self, *args):
        self.selection_clear()
        self.icursor(0)

    def end(self, *args):
        self.selection_clear()
        self.icursor('end')

    def monitor(self, *args):
        self.state = self.getstate()

    def change(self, *args):
        if not self.changed:
            self.laststate = self.state
            self.changed = 1
            self['foreground'] = self.editingforeground

    def cancel(self, *args):
        if self.changed:
            self.variable.set(self.value)
            self['foreground'] = self.foreground
            self.setstate(self.laststate)
            self.changed = 0

    def confirm(self, *args):
        if self.changed:
            newvalue = self.variable.get()
            if self.command:
                self.command(self, self.value, newvalue)
            self['foreground'] = self.foreground
            self.value = newvalue
            self.changed = 0

    def get(self):
        return self.value

    def set(self, value):
        self.variable.set(value)
        self.value = value
        self['foreground'] = self.foreground
        self.changed = 0
        self.laststate = self.state = len(value), None
        self.setstate(self.state)

class Toggle(Frame):
    def __init__(self, parent, command=None):
        Frame.__init__(self, parent, borderwidth=1, relief=RAISED)
        self.label = Label(self, font='{Zapf Dingbats} 10', borderwidth=2)
        self.label.pack()
        self.label.bind('<Enter>', self.enter)
        self.label.bind('<Leave>', self.leave)
        self.label.bind('<Button>', self.click)
        self.pushed = self.inside = 0
        self.command = command
        self.update()

    def update(self):
        if self.pushed:
            self.label['text'] = 's' # triangle pointing up
            self['relief'] = SUNKEN
            self.label['background'] = self.inside and '#e0e0e0' or '#c0c0c0'
        else:
            self.label['text'] = 't' # triangle pointing down
            self['relief'] = RAISED
            self.label['background'] = self.inside and '#ffffff' or '#e0e0e0'

    def click(self, *args):
        self.pushed = not self.pushed
        self.update()
        if self.command:
            self.command(self.pushed)

    def enter(self, *args):
        self.inside = 1
        self.update()

    def leave(self, *args):
        self.inside = 0
        self.update()
